from typing import NamedTuple, Sequence
import asyncio
import copy

import kr8s

import celpy
from celpy import celtypes

from koreo import result
from koreo.cel.evaluation import evaluate
from koreo.conditions import Condition
from koreo.resource_function.reconcile import reconcile_resource_function
from koreo.value_function.reconcile import reconcile_value_function

from . import structure

# TODO: What is reasonable here? Perhaps 10 seconds?
STEP_TIMEOUT = 10
TIMEOUT_RETRY_DELAY = 30
UNKNOWN_ERROR_RETRY_DELAY = 60


ResourceIds = dict | list["ResourceIds"] | None


class Result(NamedTuple):
    result: result.UnwrappedOutcome[celtypes.Value]
    conditions: list[Condition]
    resource_ids: dict[str, ResourceIds]
    state: celtypes.MapType
    state_errors: dict[str, str]


async def reconcile_workflow(
    api: kr8s.Api,
    workflow_key: str,
    owner: tuple[str, dict],
    trigger: celtypes.Value,
    workflow: structure.Workflow,
) -> Result:
    # This should block no-steps and any non-ok steps.
    # TODO: Make sure to handle resource-ids so we don't cause a problem
    if not result.is_ok(workflow.steps_ready):
        updated_outcome = copy.deepcopy(workflow.steps_ready)
        updated_outcome.location = f"{workflow_key}:{updated_outcome.location}"
        condition = _condition_helper(
            condition_type="Ready",
            thing_name=f"Workflow {workflow_key}",
            outcome=updated_outcome,
            workflow_key=workflow_key,
        )
        return Result(
            result=updated_outcome,
            conditions=[condition],
            resource_ids={},
            state=celtypes.MapType({}),
            state_errors={},
        )

    outcomes, conditions, step_state, state_errors = await _reconcile_steps(
        api=api,
        workflow_key=workflow_key,
        steps=workflow.steps,
        owner=owner,
        trigger=trigger,
    )

    outcome_results = {key: result.result for key, result in outcomes.items()}

    outcome_resources = {
        "workflow": workflow.name,
        "resources": {key: result.resource_ids for key, result in outcomes.items()},
    }

    overall_outcome = result.unwrapped_combine(outcomes=outcome_results.values())
    conditions.append(
        _condition_helper(
            condition_type="Ready",
            thing_name=f"Workflow {workflow_key}",
            outcome=overall_outcome,
            workflow_key=workflow_key,
        )
    )

    return Result(
        result=overall_outcome,
        conditions=conditions,
        resource_ids=outcome_resources,
        state=step_state,
        state_errors=state_errors,
    )


class StepResult(NamedTuple):
    result: result.UnwrappedOutcome[celtypes.Value]
    resource_ids: ResourceIds = None


async def _reconcile_steps(
    api: kr8s.Api,
    workflow_key: str,
    owner: tuple[str, dict],
    trigger: celtypes.Value,
    steps: Sequence[structure.Step | structure.ErrorStep],
) -> tuple[dict[str, StepResult], list[Condition], celtypes.MapType, dict[str, str]]:
    if not steps:
        return {}, [], celtypes.MapType({}), {}

    outcome_map: dict[
        str, tuple[structure.StepConditionSpec | None, celpy.Runner | None]
    ] = {}
    task_map: dict[str, asyncio.Task[StepResult]] = {}

    try:
        async with asyncio.timeout(STEP_TIMEOUT), asyncio.TaskGroup() as task_group:
            if steps:
                for step in steps:
                    match step:
                        case structure.Step(dynamic_input_keys=dynamic_input_keys):
                            step_dependencies = [
                                task_map[dependency]
                                for dependency in dynamic_input_keys
                            ]

                        case structure.ErrorStep():
                            # TODO: Could probably short-circuit here and set step.outcome?
                            step_dependencies = []

                    task_map[step.label] = task_group.create_task(
                        _reconcile_step(
                            api=api,
                            workflow_key=workflow_key,
                            step=step,
                            owner=owner,
                            trigger=trigger,
                            dependencies=step_dependencies,
                        ),
                        name=step.label,
                    )
                    outcome_map[step.label] = (step.condition, step.state)
    except:
        # Exceptions will be processed for each task
        pass

    outcomes: dict[str, StepResult] = {}
    conditions = []
    state = celtypes.MapType({})
    state_errors: dict[str, str] = {}

    for task in task_map.values():
        task_name = task.get_name()

        if task.cancelled():
            timeout_outcome = StepResult(
                result=result.Retry(
                    message=f"Timeout running step ({task_name}), will retry.",
                    delay=TIMEOUT_RETRY_DELAY,
                    location=workflow_key,
                )
            )
            outcomes[task_name] = timeout_outcome
            conditions.append(
                _condition_helper(
                    condition_type="Ready",
                    thing_name=f"Workflow {workflow_key}",
                    outcome=timeout_outcome,
                    workflow_key=workflow_key,
                )
            )

        elif task.exception():
            error_outcome = StepResult(
                result=result.Retry(
                    message=f"Unknown error ({task.exception()}) running Step ({task_name}), will retry.",
                    delay=UNKNOWN_ERROR_RETRY_DELAY,
                    location=workflow_key,
                )
            )
            outcomes[task_name] = error_outcome
            conditions.append(
                _condition_helper(
                    condition_type="Ready",
                    thing_name=f"Workflow {workflow_key}",
                    outcome=error_outcome,
                    workflow_key=workflow_key,
                )
            )

        elif task.done():
            step_result = task.result()
            outcomes[task_name] = step_result
            condition_config, state_runner = outcome_map[task_name]
            if condition_config:
                conditions.append(
                    _condition_helper(
                        condition_type=condition_config.type_,
                        thing_name=condition_config.name,
                        outcome=step_result.result,
                        workflow_key=workflow_key,
                    )
                )
            if state_runner and result.is_unwrapped_ok(step_result.result):
                match evaluate(
                    expression=state_runner,
                    inputs={"value": step_result.result},
                    location=f"{task_name}:state",
                ):
                    case celtypes.MapType() as step_state:
                        state.update(step_state)
                    case result.PermFail() as err:
                        state_errors[task_name] = f"{err.message} at {err.location}"
                    case _ as bad_state:
                        state_errors[task_name] = (
                            f"Invalid state type ({type(bad_state)})"
                        )

    return outcomes, conditions, state, state_errors


def _outcome_encoder(outcome: result.UnwrappedOutcome):
    if result.is_error(outcome):
        return outcome

    if result.is_skip(outcome):
        # This will be Skip or DepSkip, which are informational.
        # TODO: Should these be encoded some other way?
        return celtypes.StringType(f"{outcome}")

    # This should be an unwrapped-Ok value
    return outcome


async def _reconcile_step(
    api: kr8s.Api,
    workflow_key: str,
    step: structure.Step | structure.ErrorStep,
    trigger: celtypes.Value,
    dependencies: list[asyncio.Task[StepResult]],
    owner: tuple[str, dict],
) -> StepResult:
    location = f"{workflow_key}.spec.steps.{step.label}"

    if isinstance(step, structure.ErrorStep):
        return StepResult(result=step.outcome)

    if not dependencies:
        workflow_inputs = celtypes.MapType({"parent": trigger})

        if not step.inputs:
            inputs = celtypes.MapType()
        else:
            inputs = evaluate(
                expression=step.inputs,
                inputs=workflow_inputs,
                location=f"{location}.inputs",
            )
            if not result.is_unwrapped_ok(inputs):
                return StepResult(result=inputs)

        if step.skip_if:
            match evaluate(
                expression=step.skip_if,
                inputs=workflow_inputs,
                location=f"{location}.skipIf",
            ):
                case result.PermFail() as err:
                    return StepResult(result=err)
                case celtypes.BoolType() as should_skip:
                    if should_skip:
                        return StepResult(
                            result=result.Skip(message="skipped by workflow")
                        )
                case _ as bad_type:
                    return StepResult(
                        result=result.PermFail(
                            message=f"`skipIf` must evaluate to a bool, received '{bad_type}' a {type(bad_type)}",
                            location=f"{location}.skipIf",
                        )
                    )

        if step.for_each:
            return await _for_each_reconciler(
                api=api,
                workflow_key=workflow_key,
                location=location,
                step=step,
                owner=owner,
                workflow_inputs=workflow_inputs,
                inputs=inputs,
            )
        else:
            return await _reconcile_step_logic(
                api=api,
                workflow_key=workflow_key,
                location=location,
                logic=step.logic,
                owner=owner,
                workflow_inputs=workflow_inputs,
                inputs=inputs,
            )

    # TODO: Error handling review
    _, pending = await asyncio.wait(dependencies)
    if pending:
        timed_out_tasks = ", ".join(task.get_name() for task in pending)
        return StepResult(
            result=result.Retry(
                message=f"Timeout running Workflow Steps ({timed_out_tasks}), will retry.",
                delay=15,
                location=location,
            )
        )

    ok_outcomes = celtypes.MapType()

    for task in dependencies:
        step_label = task.get_name()
        step_result = task.result()
        match step_result.result:
            case result.DepSkip(message=skip_message, location=skip_location):
                return StepResult(
                    result=result.DepSkip(
                        f"'{step_label}' is waiting on dependency ({skip_message} at {skip_location}).",
                        location=location,
                    )
                )

            case result.Skip(message=skip_message, location=skip_location):
                return StepResult(
                    result=result.DepSkip(
                        f"'{step_label}' was skipped ({skip_message} at {skip_location}).",
                        location=location,
                    )
                )

            case result.Retry(message=retry_message, location=retry_location):
                return StepResult(
                    result=result.DepSkip(
                        f"'{step_label}' is waiting ({retry_message} at {retry_location}).",
                        location=location,
                    )
                )

            case result.PermFail(message=fail_message, location=fail_location):
                return StepResult(
                    result=result.DepSkip(
                        f"'{step_label}' is in failure state ({fail_message} at {fail_location}).",
                        location=location,
                    )
                )

            case result.Ok(data=data):
                ok_outcomes[celtypes.StringType(step_label)] = data

            case _ as data:
                # This should be an UnwrappedOutcome (Ok)
                ok_outcomes[celtypes.StringType(step_label)] = data

    workflow_inputs: dict[str, celtypes.Value] = celtypes.MapType(
        {
            "steps": ok_outcomes,
            "parent": trigger,
        }
    )
    if not step.inputs:
        inputs = celtypes.MapType()
    else:
        inputs = evaluate(
            expression=step.inputs,
            inputs=workflow_inputs,
            location=f"{location}.inputs",
        )
        if not result.is_unwrapped_ok(inputs):
            return StepResult(result=inputs)

    if step.skip_if:
        match evaluate(
            expression=step.skip_if,
            inputs=workflow_inputs,
            location=f"{location}.skipIf",
        ):
            case result.PermFail() as err:
                return StepResult(result=err)
            case celtypes.BoolType() as should_skip:
                if should_skip:
                    return StepResult(result=result.Skip(message="skipped by workflow"))
            case _ as bad_type:
                return StepResult(
                    result=result.PermFail(
                        message=f"`skipIf` must evaluate to a bool, received '{bad_type}' a {type(bad_type)}",
                        location=f"{location}.skipIf",
                    )
                )

    if step.for_each:
        return await _for_each_reconciler(
            api=api,
            workflow_key=workflow_key,
            location=location,
            step=step,
            workflow_inputs=workflow_inputs,
            owner=owner,
            inputs=inputs,
        )
    else:
        return await _reconcile_step_logic(
            api=api,
            workflow_key=workflow_key,
            location=location,
            logic=step.logic,
            owner=owner,
            inputs=inputs,
            workflow_inputs=workflow_inputs,
        )


async def _reconcile_ref_switch(
    api: kr8s.Api,
    workflow_key: str,
    owner: tuple[str, dict],
    logic_switch: structure.LogicSwitch,
    workflow_inputs: celtypes.MapType,
    inputs: celtypes.Value,
    location: str,
):

    # This gives the switch-on expression access to direct outcomes through
    # `steps`, but also to the step's `inputs`. In addition to possible
    # convenience, this gives the switch access to the iterated values from a
    # for-each expansion.
    switch_on_inputs: dict[str, celtypes.Value] = {"inputs": inputs} | workflow_inputs

    match evaluate(
        expression=logic_switch.switch_on,
        inputs=switch_on_inputs,
        location=f"{location}.switchOn",
    ):
        case result.PermFail() as err:
            return StepResult(result=err)
        case (celtypes.StringType() | celtypes.IntType()) as switch_value:
            # Just needed to extract the `switch_on` value
            pass
        case bad_type:
            return StepResult(
                result=result.PermFail(
                    message=f"`switchOn` must evaluate to a string or int value, received '{bad_type}' a {type(bad_type)}",
                    location=f"{location}.switchOn",
                )
            )

    logic = logic_switch.logic_map.get(switch_value, logic_switch.default_logic)

    if not logic:
        # TODO: Should this be a PermFail or a Retry? I _think_ PermFail
        # because it really should be guarded by a `skipIf` if there's a
        # non-ready condition.
        return StepResult(
            result=result.PermFail(
                message=(
                    "`refSwitch` must match one case or a default must be provided, "
                    f"failed to match '{switch_value}' a {type(switch_value)}. "
                    "Use a `skipIf` guard if this is expected."
                ),
                location=f"{location}.<match-case>",
            )
        )

    return await _reconcile_step_logic(
        api=api,
        workflow_key=workflow_key,
        owner=owner,
        inputs=inputs,
        workflow_inputs=workflow_inputs,
        location=f"{location}['{switch_value}']",
        logic=logic,
    )


async def _reconcile_step_logic(
    api: kr8s.Api,
    workflow_key: str,
    owner: tuple[str, dict],
    inputs: celtypes.Value,
    workflow_inputs: celtypes.MapType,
    location: str,
    logic: (
        structure.ResourceFunction
        | structure.ValueFunction
        | structure.Workflow
        | structure.LogicSwitch
        | result.NonOkOutcome
    ),
) -> StepResult:
    match logic:
        case structure.Workflow():
            workflow_result = await reconcile_workflow(
                api=api,
                workflow_key=workflow_key,
                owner=owner,
                trigger=inputs,
                workflow=logic,
            )
            if result.is_unwrapped_ok(workflow_result.result):
                return StepResult(
                    result=workflow_result.state,
                    resource_ids=workflow_result.resource_ids,
                )

            return StepResult(
                result=workflow_result.result, resource_ids=workflow_result.resource_ids
            )

        case structure.ResourceFunction():
            func_result, resource_id = await reconcile_resource_function(
                api=api,
                location=location,
                function=logic,
                owner=owner,
                inputs=inputs,
            )
            return StepResult(result=func_result, resource_ids=resource_id)

        case structure.ValueFunction():
            return StepResult(
                result=await reconcile_value_function(
                    location=location,
                    function=logic,
                    inputs=inputs,
                )
            )

        case structure.LogicSwitch():
            return await _reconcile_ref_switch(
                api=api,
                workflow_key=workflow_key,
                owner=owner,
                logic_switch=logic,
                workflow_inputs=workflow_inputs,
                inputs=inputs,
                location=f"{location}.refSwitch",
            )

    return StepResult(result=logic)


async def _for_each_reconciler(
    api: kr8s.Api,
    step: structure.Step,
    workflow_key: str,
    location: str,
    owner: tuple[str, dict],
    workflow_inputs: celtypes.MapType,
    inputs: celtypes.Value,
) -> StepResult:
    assert step.for_each

    match evaluate(
        expression=step.for_each.source_iterator,
        inputs=workflow_inputs,
        location=f"{step.label}.forEach.itemIn",
    ):
        case result.PermFail() as failure:
            return StepResult(result=failure)

        case celtypes.ListType() as source_iterator:
            # Without this, if the source_iterator is empty the task-group
            # raises an exception at context-manager exit.
            if not source_iterator:
                return StepResult(result=celtypes.ListType())

        case _ as bad_type:
            return StepResult(
                result=result.PermFail(
                    message=f"Step `{location}.forEach.itemIn` must be a list-type, received {type(bad_type)}.",
                    location=location,
                )
            )

    tasks: list[asyncio.Task[StepResult]] = []

    try:
        async with asyncio.timeout(STEP_TIMEOUT), asyncio.TaskGroup() as task_group:
            for idx, map_value in enumerate(source_iterator):
                iterated_inputs = copy.deepcopy(inputs)
                iterated_inputs[step.for_each.input_key] = map_value
                tasks.append(
                    task_group.create_task(
                        _reconcile_step_logic(
                            api=api,
                            workflow_key=workflow_key,
                            location=f"{location}[{idx}]",
                            logic=step.logic,
                            owner=owner,
                            inputs=iterated_inputs,
                            workflow_inputs=workflow_inputs,
                        ),
                        name=f"{step.label}-{idx}",
                    )
                )
    except:
        # Handled per-task below
        pass

    # TODO: This shouldn't even be needed. The context manager does this already?
    _, pending = await asyncio.wait(tasks)
    if pending:
        timed_out_tasks = ", ".join(task.get_name() for task in pending)
        return StepResult(
            result=result.Retry(
                message=f"Timeout running Workflow For Each Step ({timed_out_tasks}), will retry.",
                delay=15,
                location=location,
            )
        )

    outcomes = []
    for task in tasks:
        task_name = task.get_name()

        if task.cancelled():
            outcomes.append(
                StepResult(
                    result=result.Retry(
                        message=f"Timeout running step ({task_name}), will retry.",
                        delay=TIMEOUT_RETRY_DELAY,
                        location=workflow_key,
                    )
                )
            )

        elif task.exception():
            outcomes.append(
                StepResult(
                    result=result.Retry(
                        message=f"Error ({task.exception()}) running Step ({task_name}), will retry.",
                        delay=UNKNOWN_ERROR_RETRY_DELAY,
                        location=workflow_key,
                    )
                )
            )

        elif task.done():
            outcomes.append(task.result())

        else:
            outcomes.append(
                StepResult(
                    result=result.Retry(
                        message=f"Unexpected task-state running Step ({task_name}), will retry.",
                        delay=UNKNOWN_ERROR_RETRY_DELAY,
                        location=workflow_key,
                    )
                )
            )

    error_outcome = result.combine(
        [outcome.result for outcome in outcomes if result.is_error(outcome.result)]
    )
    resource_ids = [outcome.resource_ids for outcome in outcomes]

    if result.is_error(error_outcome):
        return StepResult(result=error_outcome, resource_ids=resource_ids)

    return StepResult(
        result=celtypes.ListType(
            [_outcome_encoder(outcome.result) for outcome in outcomes]
        ),
        resource_ids=resource_ids,
    )


def _condition_helper(
    condition_type: str,
    thing_name: str,
    outcome: result.UnwrappedOutcome | None,
    workflow_key: str,
) -> Condition:
    reason = "Pending"
    message: str = f"Awaiting {thing_name} reconciliation."
    status = "True"
    location = workflow_key

    if outcome is None:
        return Condition(
            type=condition_type,
            reason=reason,
            message=message,
            status=status,
            location=location,
        )

    match outcome:
        case result.DepSkip(message=skip_message, location=location):
            reason = "DepSkip"
            message: str = f"{thing_name} awaiting dependencies to be ready."
            if skip_message:
                message: str = (
                    f"{thing_name} awaiting dependencies to be ready ({skip_message})"
                )

        case result.Skip(message=skip_message, location=location):
            reason = "Skip"
            message: str = f"{thing_name} management is disabled."
            if skip_message:
                message: str = f"Skipping {thing_name}: {skip_message}"

        case result.Retry(message=retry_message, location=location):
            reason = "Wait"
            message: str = f"Awaiting {thing_name} to be ready."
            if retry_message:
                message: str = f"Awaiting {thing_name} to be ready ({retry_message})."

        case result.PermFail(message=fail_message, location=location):
            reason = "Failure"
            message: str = (
                f"Unrecoverable error reconciling {thing_name}. ({fail_message})."
            )

        case result.Ok(location=location):
            reason = "Ready"
            message: str = f"{thing_name} ready."

        case _:
            # This should be an UnwrappedOutcome (Ok)
            reason = "Ready"
            message: str = f"{thing_name} ready."

    return Condition(
        type=condition_type,
        reason=reason,
        message=message,
        status=status,
        location=location,
    )



import celpy
from celpy import celtypes


from koreo.cel.evaluation import evaluate, evaluate_predicates, evaluate_overlay
from koreo.result import PermFail, UnwrappedOutcome

from .structure import ValueFunction


async def reconcile_value_function(
    location: str,
    function: ValueFunction,
    inputs: celtypes.Value,
    value_base: celtypes.MapType | None = None,
) -> UnwrappedOutcome[celtypes.Value]:
    full_inputs: dict[str, celtypes.Value] = {
        "inputs": inputs,
    }

    # TODO: Need to decide if this is correct to do.
    if value_base:
        full_inputs = full_inputs | {celtypes.StringType("resource"): value_base}

    if precondition_error := evaluate_predicates(
        predicates=function.preconditions,
        inputs=full_inputs,
        location=f"{location}:spec.preconditions",
    ):
        return precondition_error

    # If there's no return_value, just bail out. No point in extra work.
    if not function.return_value:
        return celpy.json_to_cel(None)

    match evaluate(
        expression=function.local_values,
        inputs=full_inputs,
        location=f"{location}:spec.locals",
    ):
        case PermFail() as err:
            return err
        case celtypes.MapType() as local_values:
            full_inputs["locals"] = local_values
        case None:
            full_inputs["locals"] = celtypes.MapType({})
        case bad_type:
            # Due to validation within `prepare`, this should never happen.
            return PermFail(
                message=f"Invalid `locals` expression type ({type(bad_type)})",
                location=f"{location}:spec.locals",
            )

    if value_base is None:
        value_base = celtypes.MapType({})

    return evaluate_overlay(
        overlay=function.return_value,
        inputs=full_inputs,
        base=value_base,
        location=f"{location}:spec.return",
    )



import unittest

from celpy import celtypes

from koreo.result import PermFail, is_unwrapped_ok

from koreo.resource_template.prepare import prepare_resource_template


class TestPrepareResourceTemplate(unittest.IsolatedAsyncioTestCase):
    async def test_missing_spec(self):
        prepared = await prepare_resource_template("test-case", {})
        self.assertIsInstance(prepared, PermFail)
        self.assertIn("spec must contain", prepared.message)

    async def test_missing_template(self):
        prepared = await prepare_resource_template(
            "test-case",
            {"context": {}},
        )
        self.assertIsInstance(prepared, PermFail)
        self.assertIn("must contain", prepared.message)

    async def test_bad_template_type(self):
        prepared = await prepare_resource_template(
            "test-case",
            {
                "template": "bad value",
            },
        )
        self.assertIsInstance(prepared, PermFail)
        self.assertIn("spec.template must be object", prepared.message)

    async def test_missing_apiVersion(self):
        prepared = await prepare_resource_template(
            "test-case",
            {
                "template": {
                    "kind": "TestResource",
                },
            },
        )
        self.assertIsInstance(prepared, PermFail)
        self.assertIn("`apiVersion` and `kind` must be set", prepared.message)

    async def test_bad_context(self):
        prepared = await prepare_resource_template(
            "test-case",
            {
                "template": {
                    "apiVersion": "api.group/v1",
                    "kind": "TestResource",
                },
                "context": "bad value",
            },
        )
        self.assertIsInstance(prepared, PermFail)
        self.assertIn("must be object", prepared.message)

    async def test_good_config(self):
        prepared = await prepare_resource_template(
            "test-case",
            {
                "template": {
                    "apiVersion": "api.group/v1",
                    "kind": "TestResource",
                    "spec": {"bool": True},
                },
                "context": {"values": "ok"},
            },
        )

        self.assertTrue(is_unwrapped_ok(prepared))
        prepared_template, _ = prepared

        self.assertIsInstance(
            prepared_template.template.get("spec", {}).get("bool"), celtypes.BoolType
        )



import unittest

from koreo.result import DepSkip, is_error, is_unwrapped_ok

from koreo.function_test import prepare


class TestMalformedFunctionTests(unittest.IsolatedAsyncioTestCase):
    async def test_missing_spec(self):
        # None
        outcome = await prepare.prepare_function_test(cache_key="unit-test", spec=None)
        self.assertTrue(is_error(outcome))

        # Empty dict
        outcome = await prepare.prepare_function_test(cache_key="unit-test", spec={})
        self.assertTrue(is_error(outcome))

    async def test_missing_function(self):
        outcome = await prepare.prepare_function_test(
            cache_key="unit-test",
            spec={
                "functionRef": {
                    "kind": "ValueFunction",
                    "name": "unit-test-function.fake",
                },
                "testCases": [{"expectOutcome": {"ok": {}}}],
            },
        )
        assert is_unwrapped_ok(outcome), outcome.message

        function_test, _ = outcome
        self.assertIsInstance(function_test.function_under_test, DepSkip)



from typing import Sequence
import json
import logging

from celpy import celtypes
import celpy

from koreo import result
from koreo.cel.encoder import encode_cel
from koreo.cel.functions import koreo_cel_functions


def predicate_extractor(
    cel_env: celpy.Environment,
    predicate_spec: Sequence[dict] | None,
) -> None | celpy.Runner | result.PermFail:
    if not predicate_spec:
        return None

    if not isinstance(predicate_spec, (list, tuple)):
        return result.PermFail(message="Malformed conditions, expected a list")

    predicates = encode_cel(predicate_spec)
    conditions = f"{predicates}.filter(predicate, !predicate.assert)"

    try:
        program = cel_env.program(
            cel_env.compile(conditions), functions=koreo_cel_functions
        )
    except celpy.CELParseError as err:
        return result.PermFail(
            message=f"Parsing error at line {err.line}, column {err.column}. "
            f"'{err}' in '{predicates}'",
        )

    program.logger.setLevel(logging.WARNING)
    return program


def predicate_to_koreo_result(
    predicates: celtypes.ListType, location: str
) -> result.NonOkOutcome | None:
    if not predicates:
        return None

    for predicate in predicates:
        match predicate:
            case {"assert": _, "ok": {}}:
                return None

            case {"assert": _, "depSkip": {"message": message}}:
                return result.DepSkip(message=f"{message}", location=location)

            case {"assert": _, "skip": {"message": message}}:
                return result.Skip(message=f"{message}", location=location)

            case {"assert": _, "retry": {"message": message, "delay": delay}}:
                return result.Retry(
                    message=f"{message}",
                    delay=int(f"{delay}"),
                    location=location,
                )

            case {"assert": _, "permFail": {"message": message}}:
                return result.PermFail(message=f"{message}", location=location)

            case _:
                return result.PermFail(
                    f"Unknown predicate type: {json.dumps(predicate)}",
                    location=location,
                )

    return None



import unittest

import celpy

from koreo.cel.structure_extractor import extract_argument_structure


class TestArgumentStructureExtractor(unittest.TestCase):
    def test_simple_args(self):
        env = celpy.Environment()

        cel_str = """{
            "simple_dot": steps.simple1,
            "simple_index": steps['simple2'],
            "nested_dots": steps.nested1.one,
            "nested_index_dot": steps['nested2'].two,
            "nested_index_index": steps['nested3']['one'],
            "nested_dot_index": steps.nested4['one']
        }
        """

        result = extract_argument_structure(env.compile(cel_str))

        sorted_result = sorted(result)

        expected = sorted(
            [
                "steps.nested1",
                "steps.nested2",
                "steps.nested3",
                "steps.nested4",
                "steps.simple1",
                "steps.simple2",
                "steps.nested1.one",
                "steps.nested2.two",
                "steps.nested3.one",
                "steps.nested4.one",
            ]
        )

        self.assertListEqual(expected, sorted_result)

    def test_numeric_indexes(self):
        env = celpy.Environment()

        cel_str = """{
            "simple_index": steps[0],
            "nested_index_dot": steps[2].two,
            "nested_index_index": steps[3][1],
            "nested_dot_index": steps.nested4[4]
        }
        """

        result = extract_argument_structure(env.compile(cel_str))

        sorted_result = sorted(result)

        expected = sorted(
            [
                "steps.0",
                "steps.2",
                "steps.2.two",
                "steps.3",
                "steps.3.1",
                "steps.nested4",
                "steps.nested4.4",
            ]
        )

        self.assertListEqual(expected, sorted_result)

    def test_formula_args(self):
        env = celpy.Environment()

        cel_str = """{
            "simple_add": steps.simple1 + steps['simple2'],
            "nested_add": steps.nested1.one + steps['nested2'].two + steps['nested3']['one'] + steps.nested4['one']
        }
        """

        result = extract_argument_structure(env.compile(cel_str))

        sorted_result = sorted(result)

        expected = sorted(
            [
                "steps.nested1",
                "steps.nested2",
                "steps.nested3",
                "steps.nested4",
                "steps.simple1",
                "steps.simple2",
                "steps.nested1.one",
                "steps.nested2.two",
                "steps.nested3.one",
                "steps.nested4.one",
            ]
        )

        self.assertListEqual(expected, sorted_result)



import logging

logger = logging.getLogger("koreo.valuefunction.prepare")

import celpy

from koreo import schema
from koreo.cel.functions import koreo_function_annotations
from koreo.cel.prepare import (
    Overlay,
    prepare_map_expression,
    prepare_overlay_expression,
)
from koreo.cel.structure_extractor import extract_argument_structure
from koreo.predicate_helpers import predicate_extractor
from koreo.result import PermFail, UnwrappedOutcome

from . import structure

# Try to reduce the incredibly verbose logging from celpy
logging.getLogger("Environment").setLevel(logging.WARNING)
logging.getLogger("NameContainer").setLevel(logging.WARNING)
logging.getLogger("Evaluator").setLevel(logging.WARNING)
logging.getLogger("evaluation").setLevel(logging.WARNING)
logging.getLogger("celtypes").setLevel(logging.WARNING)


async def prepare_value_function(
    cache_key: str, spec: dict
) -> UnwrappedOutcome[tuple[structure.ValueFunction, None]]:
    logger.debug(f"Prepare ValueFunction:{cache_key}")

    if error := schema.validate(
        resource_type=structure.ValueFunction, spec=spec, validation_required=True
    ):
        return PermFail(
            error.message,
            location=_location(cache_key, "spec"),
        )

    env = celpy.Environment(annotations=koreo_function_annotations)

    used_vars = set[str]()

    match predicate_extractor(cel_env=env, predicate_spec=spec.get("preconditions")):
        case PermFail(message=message):
            return PermFail(
                message=message, location=_location(cache_key, "spec.preconditions")
            )
        case None:
            preconditions = None
        case celpy.Runner() as preconditions:
            used_vars.update(extract_argument_structure(preconditions.ast))

    match prepare_map_expression(
        cel_env=env, spec=spec.get("locals"), location="spec.locals"
    ):
        case PermFail(message=message):
            return PermFail(
                message=message,
                location=_location(cache_key, "spec.locals"),
            )
        case None:
            local_values = None
        case celpy.Runner() as local_values:
            used_vars.update(extract_argument_structure(local_values.ast))

    match prepare_overlay_expression(
        cel_env=env, spec=spec.get("return"), location="spec.return"
    ):
        case PermFail(message=message):
            return PermFail(
                message=message,
                location=_location(cache_key, "spec.return"),
            )
        case None:
            return_value = None

        case Overlay(values=values) as return_value:
            used_vars.update(extract_argument_structure(values.ast))

    return (
        structure.ValueFunction(
            preconditions=preconditions,
            local_values=local_values,
            return_value=return_value,
            dynamic_input_keys=used_vars,
        ),
        None,
    )


def _location(cache_key: str, extra: str | None = None) -> str:
    base = f"prepare:ValueFunction:{cache_key}"
    if not extra:
        return base

    return f"{base}:{extra}"



import unittest

from koreo import cache
from koreo import result
from koreo.resource_function.prepare import prepare_resource_function
from koreo.resource_function.structure import ResourceFunction

from koreo.function_test import prepare
from koreo.function_test import run


class TestEndToEndResourceFunction(unittest.IsolatedAsyncioTestCase):
    async def asyncTearDown(self):
        if cache.get_resource_from_cache(
            resource_class=ResourceFunction, cache_key="function-under-test"
        ):
            await cache.delete_from_cache(
                resource_class=ResourceFunction, cache_key="function-under-test"
            )

    async def asyncSetUp(self):
        if cache.get_resource_from_cache(
            resource_class=ResourceFunction, cache_key="function-under-test"
        ):
            cache.delete_from_cache
            return

        await cache.prepare_and_cache(
            resource_class=ResourceFunction,
            preparer=prepare_resource_function,
            metadata={"name": "function-under-test", "resourceVersion": "sam123"},
            spec={
                "apiConfig": {
                    "apiVersion": "test.koreo.dev/v1",
                    "kind": "TestResource",
                    "name": "=inputs.metadata.name + '-suffix'",
                    "namespace": "=inputs.metadata.namespace",
                },
                "preconditions": [
                    {
                        "assert": "=!has(inputs.preconditions.depSkip)",
                        "depSkip": {"message": "Input Validator Dep Skip"},
                    },
                    {
                        "assert": "=!has(inputs.preconditions.skip)",
                        "skip": {"message": "Input Validator Skip"},
                    },
                    {
                        "assert": "=!has(inputs.preconditions.permFail)",
                        "permFail": {"message": "Input Validator Perm Fail"},
                    },
                    {
                        "assert": "=!has(inputs.preconditions.retry)",
                        "retry": {"delay": 18, "message": "Input Validator Retry"},
                    },
                ],
                "resource": {
                    "apiVersion": "test.koreo.dev/v1",
                    "kind": "TestResource",
                    "metadata": "=inputs.metadata",
                    "spec": {
                        "subStructure": {
                            "int": 1,
                            "bool": True,
                            "list": [1, 2, 3],
                            "object": {"one": 1, "two": 2},
                        }
                    },
                },
                "create": {
                    "overlay": {
                        "spec": {
                            "subStructure": {
                                "createOnly": "value",
                            },
                        },
                    },
                },
                "postconditions": [
                    {
                        "assert": "=!has(resource.status.postconditions.depSkip)",
                        "depSkip": {"message": "Output Validator Dep Skip"},
                    },
                    {
                        "assert": "=!has(resource.status.postconditions.skip)",
                        "skip": {"message": "Output Validator Skip"},
                    },
                    {
                        "assert": "=!has(resource.status.postconditions.permFail)",
                        "permFail": {"message": "Output Validator Perm Fail"},
                    },
                    {
                        "assert": "=!has(resource.status.postconditions.retry)",
                        "retry": {"delay": 18, "message": "Output Validator Retry"},
                    },
                    {
                        "assert": "=has(resource.status.returnValue)",
                        "retry": {"delay": 1, "message": "Waiting for ready-state"},
                    },
                ],
                "return": {"value": "=resource.status.returnValue"},
            },
        )

    async def test_resource_function_fully(self):
        prepared_test = await prepare.prepare_function_test(
            cache_key="function-test:end-to-end",
            spec={
                "functionRef": {
                    "kind": "ResourceFunction",
                    "name": "function-under-test",
                },
                "inputs": {
                    "metadata": {
                        "name": "function-test:end-to-end",
                        "namespace": "unittests",
                    },
                },
                "testCases": [
                    {
                        "label": "Initial create",
                        "expectResource": {
                            "apiVersion": "test.koreo.dev/v1",
                            "kind": "TestResource",
                            "metadata": {
                                "name": "function-test:end-to-end-suffix",
                                "namespace": "unittests",
                            },
                            "spec": {
                                "subStructure": {
                                    "int": 1,
                                    "bool": True,
                                    "list": [1, 2, 3],
                                    "object": {"one": 1, "two": 2},
                                    "createOnly": "value",
                                }
                            },
                        },
                    },
                    {
                        "label": "[variant] Rename causes create",
                        "variant": True,
                        "inputOverrides": {
                            "metadata": {"name": "pumpkins"},
                        },
                        "expectResource": {
                            "apiVersion": "test.koreo.dev/v1",
                            "kind": "TestResource",
                            "metadata": {
                                "name": "pumpkins-suffix",
                                "namespace": "unittests",
                            },
                            "spec": {
                                "subStructure": {
                                    "int": 1,
                                    "bool": True,
                                    "list": [1, 2, 3],
                                    "object": {"one": 1, "two": 2},
                                }
                            },
                        },
                    },
                    {
                        "label": "Setting status overlay returns value",
                        "overlayResource": {
                            "status": {"returnValue": "controller-set-value"}
                        },
                        "expectReturn": {"value": "controller-set-value"},
                    },
                    {
                        "label": "Non-variant status overlay persists",
                        "expectReturn": {"value": "controller-set-value"},
                    },
                    {
                        "label": "[variant] Status overlay failure",
                        "variant": True,
                        "overlayResource": {
                            "status": {
                                "returnValue": "variant-value",
                                "postconditions": {"permFail": True},
                            }
                        },
                        "expectOutcome": {"permFail": {"message": "Output Validator"}},
                    },
                    {
                        "label": "Variant status overlay did not persist",
                        "expectReturn": {"value": "controller-set-value"},
                    },
                    {
                        "label": "[variant] Input overlay triggers preconditions",
                        "variant": True,
                        "inputOverrides": {
                            "preconditions": {"permFail": True},
                        },
                        "expectOutcome": {"permFail": {"message": "Input Validator"}},
                    },
                    {
                        "label": "variant input overlay did not persist",
                        "expectReturn": {"value": "controller-set-value"},
                    },
                ],
            },
        )

        if not result.is_unwrapped_ok(prepared_test):
            raise self.failureException(
                f"Failed to prepare function-test {prepared_test}!"
            )

        function_test, _ = prepared_test

        test_result = await run.run_function_test(
            location="unit-test:end-to-end", function_test=function_test
        )

        self.assertGreater(len(test_result.test_results), 0, "No test results")
        for idx, test_case_result in enumerate(test_result.test_results):
            if test_case_result.test_pass:
                continue

            if test_case_result.message:
                message = f"{test_case_result.message}"
            elif test_case_result.differences:
                message = f"{test_case_result.differences}"
            else:
                message = f"{test_case_result.outcome}"

            test_name = (
                test_case_result.label if test_case_result.label else f"Test {idx}"
            )

            raise self.failureException(f"{test_name} Failure: {message}")



from __future__ import annotations
from contextlib import asynccontextmanager
from typing import NamedTuple, Sequence
import copy
import json

import kr8s

import celpy
from celpy import celtypes

from koreo import result
from koreo.cel.encoder import convert_bools
from koreo.cel.evaluation import evaluate_overlay
from koreo.cel.functions import _overlay
from koreo.constants import KOREO_DIRECTIVE_KEYS, LAST_APPLIED_ANNOTATION
from koreo.resource_function.reconcile import reconcile_resource_function
from koreo.resource_function.structure import ResourceFunction
from koreo.value_function.reconcile import reconcile_value_function
from koreo.value_function.structure import ValueFunction

from .structure import (
    ExpectDelete,
    ExpectOutcome,
    ExpectResource,
    ExpectReturn,
    FunctionTest,
    TestCase,
)


class MockResponse:
    def __init__(self, data):
        self._data = data

    def json(self):
        return self._data


class MockApi:
    def __init__(self, current_resource: dict | None, *args, **kwargs):
        self._current_resource = current_resource
        self._materialized = None
        self._api_called = False
        self._delete_called = False

    @property
    def materialized(self):
        return self._materialized

    @property
    def namespace(self):
        return "FAKE-NAMESPACE"

    async def async_get(self, *args, **kwargs):
        if self._current_resource:
            # TODO: This should probably be loaded and built from the Function,
            # using _build_resource_config

            resource_class = kr8s.objects.new_class(
                version=self._current_resource.get("apiVersion"),
                kind=self._current_resource.get("kind"),
                namespaced=(
                    True
                    if self._current_resource.get("metadata", {}).get("namespace")
                    else False
                ),
                asyncio=True,
            )
            yield resource_class(
                api=self,
                resource=self._current_resource,
                namespace=self._current_resource.get("metadata", {}).get("namespace"),
            )

    @asynccontextmanager
    async def call_api(self, *args, **kwargs):
        self._api_called = True

        if args and "DELETE" in args:
            self._materialized = {}
            self._delete_called = True
            yield MockResponse(data=None)
            return

        data = json.loads(kwargs.get("data", "{}"))

        if not self._current_resource:
            merged = data
        else:
            merged = _merge_overlay(self._current_resource, data)

        self._materialized = merged
        yield MockResponse(data=merged)


def _merge_overlay(base, overlay):
    updated = copy.deepcopy(base)
    for key, value in overlay.items():
        match value:
            case dict():
                match base.get(key):
                    case dict() as sub_base:
                        updated[key] = _merge_overlay(sub_base, value)

        updated[key] = value

    return updated


class TestCaseResult(NamedTuple):
    test_pass: bool

    expected_outcome: result.Outcome | None = None
    outcome: result.UnwrappedOutcome | None = None

    message: str | None = None
    differences: Difference | None = None

    label: str | None = None


class TestResult(NamedTuple):
    test_results: Sequence[TestCaseResult]
    function_healthy: result.UnwrappedOutcome
    fatal_error: bool = False


class TestCaseOutcome(NamedTuple):
    new_inputs: celtypes.MapType | None
    new_resource: dict | None
    result: TestCaseResult
    fatal_error: bool = False


async def run_function_test(location: str, function_test: FunctionTest) -> TestResult:
    if not result.is_unwrapped_ok(function_test.function_under_test):
        return TestResult(
            test_results=[],
            function_healthy=function_test.function_under_test,
            fatal_error=True,
        )

    if not function_test.test_cases:
        return TestResult(test_results=[], function_healthy=result.Ok(None))

    test_results, fatal_error = await _run_test_cases(
        location=location,
        function_under_test=function_test.function_under_test,
        base_inputs=function_test.inputs,
        initial_resource=function_test.initial_resource,
        test_cases=function_test.test_cases,
    )
    return TestResult(
        test_results=test_results,
        function_healthy=result.Ok(None),
        fatal_error=fatal_error,
    )


async def _run_test_cases(
    location: str,
    function_under_test: ResourceFunction | ValueFunction,
    base_inputs: celtypes.MapType | None,
    initial_resource: dict | None,
    test_cases: Sequence[TestCase],
) -> tuple[Sequence[TestCaseResult], bool]:
    current_inputs = base_inputs
    current_resource = initial_resource

    test_results: list[TestCaseResult] = []
    for idx, test_case in enumerate(test_cases):
        result = await _run_test_case(
            location=location,
            current_resource=copy.deepcopy(current_resource),
            function_under_test=function_under_test,
            base_inputs=copy.deepcopy(current_inputs),
            test_case=test_case,
            idx=idx,
        )

        current_resource = result.new_resource
        current_inputs = result.new_inputs

        test_results.append(result.result)

        if result.fatal_error:
            return test_results, True

    return test_results, False


async def _run_test_case(
    location: str,
    function_under_test: ResourceFunction | ValueFunction,
    current_resource: dict | None,
    base_inputs: celtypes.MapType | None,
    test_case: TestCase,
    idx: int,
) -> TestCaseOutcome:
    location = f"{location}.testCases[{idx}]"
    if test_case.skip:
        return TestCaseOutcome(
            new_inputs=base_inputs,
            new_resource=current_resource,
            result=TestCaseResult(
                test_pass=True,
                expected_outcome=None,
                outcome=None,
                message=f"user skipped",
                label=test_case.label,
            ),
            fatal_error=False,
        )

    if base_inputs and test_case.input_overrides:
        inputs = _overlay(base_inputs, test_case.input_overrides)
        if isinstance(inputs, celpy.CELEvalError):
            return TestCaseOutcome(
                new_inputs=base_inputs,
                new_resource=current_resource,
                result=TestCaseResult(
                    test_pass=False,
                    expected_outcome=None,
                    outcome=None,
                    message=f"Could not run test case due to inputs overlay error: ({inputs})",
                    label=test_case.label,
                ),
                fatal_error=not test_case.variant,
            )

    elif base_inputs:
        inputs = base_inputs
    elif test_case.input_overrides:
        inputs = test_case.input_overrides
    else:
        inputs = celtypes.MapType()

    if test_case.resource_overlay:
        if not current_resource:
            return TestCaseOutcome(
                new_inputs=base_inputs,
                new_resource=current_resource,
                result=TestCaseResult(
                    test_pass=False,
                    expected_outcome=None,
                    outcome=None,
                    message="Can not overlay until the full resource exists. Try setting `currentResource` or reordering test cases.",
                    label=test_case.label,
                ),
                fatal_error=True,
            )

        match evaluate_overlay(
            overlay=test_case.resource_overlay,
            inputs=inputs,
            base=current_resource,
            location=f"{test_case.label}:overlayResource",
        ):
            case result.PermFail() as err:
                return TestCaseOutcome(
                    new_inputs=base_inputs,
                    new_resource=current_resource,
                    result=TestCaseResult(
                        test_pass=False,
                        expected_outcome=None,
                        outcome=err,
                        message=f"Could not run test case due to resource overlay error: ({err})",
                        label=test_case.label,
                    ),
                    fatal_error=not test_case.variant,
                )
            case _ as cel_resource:
                resource = convert_bools(cel_resource)

    elif test_case.current_resource:
        resource = test_case.current_resource
    else:
        resource = current_resource

    api = MockApi(current_resource=resource)

    match function_under_test:
        case ResourceFunction():
            reconcile_result, _ = await reconcile_resource_function(
                api=api,
                location=location,
                function=function_under_test,
                owner=(
                    f"{function_under_test.crud_config.resource_api.namespace}",
                    celtypes.MapType({"uid": "unit-test-123"}),
                ),
                inputs=inputs,
            )
        case ValueFunction():
            if not resource:
                base_resource = celtypes.MapType({})
            else:
                base_resource = celpy.json_to_cel(resource)
                assert isinstance(base_resource, celtypes.MapType)

            reconcile_result = await reconcile_value_function(
                location=location,
                function=function_under_test,
                inputs=inputs,
                value_base=base_resource,
            )

    match test_case.assertion:
        case ExpectOutcome(outcome=expected_outcome):
            outcome_result = _validate_outcome_match(
                expected=expected_outcome, actual=reconcile_result
            )
        case ExpectReturn(value=expected_value):
            outcome_result = _validate_return_match(
                expected=expected_value, actual=reconcile_result
            )
        case ExpectResource(resource=expected_resource):
            outcome_result = _validate_resource_match(
                expected=expected_resource,
                materialized=api.materialized,
                actual_outcome=reconcile_result,
            )
        case ExpectDelete(expected=expected):
            passed = api._delete_called == expected

            if api._delete_called:
                message = "Delete called"
            else:
                message = "Delete not called."

            outcome_result = TestCaseResult(
                test_pass=passed,
                expected_outcome=result.Retry(delay=0, message=""),
                outcome=reconcile_result,
                message=message,
            )

    if test_case.variant:
        return TestCaseOutcome(
            new_inputs=base_inputs,
            new_resource=current_resource,
            result=copy.replace(outcome_result, label=test_case.label),
        )

    return TestCaseOutcome(
        new_inputs=inputs,
        new_resource=api.materialized if api._api_called else resource,
        result=copy.replace(outcome_result, label=test_case.label),
        fatal_error=not outcome_result.test_pass,
    )


def _validate_return_match(
    expected: dict | None, actual: result.UnwrappedOutcome[celtypes.Value] | None
):
    if not result.is_unwrapped_ok(actual):
        return TestCaseResult(
            test_pass=False,
            expected_outcome=result.Ok(expected),
            outcome=actual,
            message=f"Non-Ok return value ({actual})",
        )

    converted_actual = convert_bools(actual)

    match = _validate_match(target=expected, actual=converted_actual)
    return TestCaseResult(
        test_pass=match.match,
        expected_outcome=result.Ok(expected),
        outcome=converted_actual,
        differences=match.differences,
    )


def _validate_resource_match(
    expected: dict | None,
    materialized: dict | None,
    actual_outcome: result.UnwrappedOutcome[celtypes.Value] | None,
):
    if materialized is None and expected is not None:
        if actual_outcome and not result.is_unwrapped_ok(actual_outcome):
            outcome_message = actual_outcome.message
        else:
            outcome_message = "unexpected Ok result"

        return TestCaseResult(
            test_pass=False,
            expected_outcome=result.Ok(expected),
            outcome=actual_outcome,
            message=f"No resource changes attempted: {outcome_message}",
        )

    if not isinstance(actual_outcome, result.Retry):
        return TestCaseResult(
            test_pass=False,
            expected_outcome=result.Ok(expected),
            outcome=actual_outcome,
            message=f"Unexpected result-type ({actual_outcome}), but expected Retry",
        )

    materialized_without_last_applied = _strip_last_applied_annotation(materialized)

    match = _validate_match(target=expected, actual=materialized_without_last_applied)
    return TestCaseResult(
        test_pass=match.match,
        expected_outcome=result.Ok(expected),
        outcome=actual_outcome,
        differences=match.differences,
    )


def _validate_outcome_match(
    expected: result.Outcome | None,
    actual: result.UnwrappedOutcome[celtypes.Value] | None,
):
    test_pass = False
    message = None

    match (expected, actual):
        case result.Retry(message=expected_message, delay=expected_delay), result.Retry(
            message=actual_message, delay=actual_delay
        ):
            if not expected_message:
                test_pass = True
            elif actual_message and (
                expected_message.lower() in actual_message.lower()
            ):
                test_pass = True

            if expected_delay and (expected_delay != actual_delay):
                # test_pass must be True here, so it needs set to False on mismatch
                test_pass = False

            if not test_pass:
                message = (
                    "Retry: "
                    f"Expected(message={expected_message}, delay={expected_delay}) "
                    f"Actual(message={actual_message}, delay={actual_delay})"
                )

        case result.PermFail(message=expected_message), result.PermFail(
            message=actual_message
        ):
            if not expected_message:
                test_pass = True
            elif actual_message and (
                expected_message.lower() in actual_message.lower()
            ):
                test_pass = True
            else:
                message = (
                    "PermFail: "
                    f"Expected(message={expected_message}) "
                    f"Actual(message={actual_message})"
                )

        case result.DepSkip(message=expected_message), result.DepSkip(
            message=actual_message
        ):
            if not expected_message:
                test_pass = True
            elif actual_message and (
                expected_message.lower() in actual_message.lower()
            ):
                test_pass = True
            else:
                message = (
                    "DepSkip: "
                    f"Expected(message={expected_message}) "
                    f"Actual(message={actual_message})"
                )

        case result.Skip(message=expected_message), result.Skip(message=actual_message):
            if not expected_message:
                test_pass = True
            elif actual_message and (
                expected_message.lower() in actual_message.lower()
            ):
                test_pass = True
            else:
                message = (
                    "Skip: "
                    f"Expected(message={expected_message}) "
                    f"Actual(message={actual_message})"
                )

        case _, _:
            if actual and result.is_unwrapped_ok(actual):
                actual = convert_bools(actual)
                actual_str = f"Ok({actual})"
            else:
                actual_str = f"{actual}"

            if expected and result.is_ok(expected):
                if not expected.data:
                    test_pass = True
                else:
                    match = _validate_match(expected.data, actual)
                    test_pass = match.match
                    if not test_pass:
                        message = f"Mismatch at {" -> ".join(match.differences)}"
            elif expected is None:
                test_pass = result.is_unwrapped_ok(actual)
                if not test_pass:
                    message = f"Expected(Ok) Actual({actual_str})"

            else:
                message = f"Expected({expected}) Actual({actual_str})"

    return TestCaseResult(
        test_pass=test_pass,
        expected_outcome=expected,
        outcome=actual,
        message=message,
    )


Difference = dict[str, "str | Difference"] | Sequence["Difference | None"] | str


class ResourceMatch(NamedTuple):
    match: bool
    differences: Difference


def _strip_last_applied_annotation[T](actual: T) -> T:
    if not actual or not isinstance(actual, dict):
        return actual

    if "metadata" not in actual:
        return actual

    if "annotations" not in actual["metadata"]:
        return actual

    annotation_count = len(actual["metadata"]["annotations"])
    if annotation_count == 0:
        return actual

    stripped = copy.deepcopy(actual)
    if annotation_count == 1:
        del stripped["metadata"]["annotations"]
    else:
        del stripped["metadata"]["annotations"][LAST_APPLIED_ANNOTATION]

    return stripped


def _validate_match(target, actual, compare_list_as_set: bool = False) -> ResourceMatch:
    """Compare the specified (`target`) state against the actual (`actual`)
    reosurce state. We compare all target fields and ignore anything extra.
    """
    match (target, actual):
        # Objects need a special comparator
        case dict(), dict():
            return _validate_dict_match(target, actual)
        case dict(), _:
            return ResourceMatch(
                match=False,
                differences=f"expected `object` but found `{type(actual)}`",
            )
        case _, dict():
            return ResourceMatch(
                match=False,
                differences=f"expected `{type(target)}` but found `object`",
            )

        # Arrays need a special comparator
        case (list() | tuple(), list() | tuple()):
            if compare_list_as_set:
                return _validate_set_match(target, actual)

            return _validate_list_match(target, actual)

        case (list() | tuple(), _):
            return ResourceMatch(
                match=False,
                differences=f"expected `array` but found `{type(actual)}`",
            )
        case (_, list() | tuple()):
            return ResourceMatch(
                match=False,
                differences=f"expected `{type(target)}` but found `array`",
            )

        # Bool needs a special comparator, due to Python's int truthiness rules
        case bool(), bool():
            if target == actual:
                return ResourceMatch(match=True, differences=())
            else:
                return ResourceMatch(
                    match=False,
                    differences=f"expected `{target}` but found `{actual}`",
                )
        case bool(), _:
            return ResourceMatch(
                match=False,
                differences=f"expected `bool` but found `{type(actual)}`",
            )
        case _, bool():
            return ResourceMatch(
                match=False,
                differences=f"expected `{type(target)}` but found `bool`",
            )

    # Hopefully anything else is a simple type.
    if target == actual:
        return ResourceMatch(match=True, differences=())

    return ResourceMatch(
        match=False,
        differences=f"expected '`{target}`' but found '`{actual}`'",
    )


def _obj_to_key(obj: dict, fields: Sequence[int | str]) -> str:
    return "$".join(f"{obj.get(field)}".strip() for field in fields)


def _list_to_object(
    obj_list: Sequence[dict], key_fields: Sequence[int | str]
) -> dict[str, dict]:
    return {_obj_to_key(obj, key_fields): obj for obj in obj_list}


def _validate_dict_match(target: dict, actual: dict) -> ResourceMatch:
    differences: dict[str, str | Difference] = {}

    compare_list_as_set_keys = {
        key for key in target.get("x-koreo-compare-as-set", ()) if key
    }

    compare_as_map = {
        key: [field_name for field_name in fields if field_name]
        for key, fields in target.get("x-koreo-compare-as-map", {}).items()
        if key
    }

    target_keys = set(target.keys()) - KOREO_DIRECTIVE_KEYS
    actual_keys = set(actual.keys()) - KOREO_DIRECTIVE_KEYS

    for missing_key in target_keys - actual_keys:
        differences[missing_key] = "missing"

    for unexpected_key in actual_keys - target_keys:
        differences[unexpected_key] = "unexpected"

    for compare_key in target_keys.intersection(actual_keys):
        if compare_key in compare_as_map:
            key_match = _validate_match(
                _list_to_object(target[compare_key], compare_as_map[compare_key]),
                _list_to_object(actual[compare_key], compare_as_map[compare_key]),
            )
        else:
            key_match = _validate_match(
                target[compare_key],
                actual[compare_key],
                compare_list_as_set=(compare_key in compare_list_as_set_keys),
            )

        if not key_match.match:
            differences[compare_key] = key_match.differences

    return ResourceMatch(match=not differences, differences=differences)


def _validate_list_match(target: list | tuple, actual: list | tuple) -> ResourceMatch:
    if len(target) != len(actual):
        return ResourceMatch(
            match=False,
            differences=f"<length mismatch target:{len(target)}, actual:{len(actual)}",
        )

    has_differences = False
    differences = []
    for target_value, actual_value in zip(target, actual):
        item_match = _validate_match(target_value, actual_value)
        if item_match.match:
            differences.append(None)
        else:
            has_differences = True
            differences.append(item_match.differences)

    return ResourceMatch(match=not has_differences, differences=differences)


def _validate_set_match(target: list | tuple, actual: list | tuple) -> ResourceMatch:
    try:
        target_set = set(target)
        actual_set = set(actual)
    except TypeError as err:
        if "dict" in f"{err}":
            return ResourceMatch(
                match=False,
                differences=f"Could not compare array-as-set, try: `x-koreo-compare-as-map`",
            )

        return ResourceMatch(
            match=False, differences=f"Could not compare array-as-set ({err})"
        )
    except Exception as err:
        return ResourceMatch(
            match=False, differences=f"Could not compare array-as-set ({err})"
        )

    missing_values = target_set - actual_set
    unexpected_values = actual_set - target_set

    if not (missing_values or unexpected_values):
        return ResourceMatch(match=True, differences=())

    differences = []

    for missing_value in missing_values:
        differences.append(
            f"<missing '{missing_value}'>",
        )

    for unexpected_value in unexpected_values:
        differences.append(
            f"<unexpectedly found '{unexpected_value}'>",
        )

    return ResourceMatch(match=False, differences=differences)



from typing import Sequence

import celpy
from celpy import celtypes

from koreo import ref_helpers
from koreo import registry
from koreo import schema
from koreo.cache import get_resource_from_cache
from koreo.cel.evaluation import evaluate
from koreo.cel.functions import _overlay, koreo_function_annotations
from koreo.cel.prepare import prepare_overlay_expression
from koreo.cel.structure_extractor import extract_argument_structure
from koreo.result import DepSkip, PermFail, UnwrappedOutcome, is_error, is_unwrapped_ok

from koreo.predicate_helpers import predicate_to_koreo_result
from koreo.resource_function.structure import ResourceFunction, ResourceTemplateRef
from koreo.resource_template.structure import ResourceTemplate

from . import structure


async def prepare_function_test(
    cache_key: str, spec: dict
) -> UnwrappedOutcome[tuple[structure.FunctionTest, Sequence[registry.Resource]]]:
    location = f"prepare:FunctionTest:{cache_key}"

    if error := schema.validate(
        resource_type=structure.FunctionTest, spec=spec, validation_required=True
    ):
        return PermFail(
            error.message,
            location=f"{cache_key}:spec",
        )

    match ref_helpers.function_ref_spec_to_resource(
        spec.get("functionRef"),
        location=f"{cache_key}:spec.functionRef",
    ):
        case PermFail() as perm_fail:
            return perm_fail
        case None:
            return PermFail(
                "Missing `spec.functionRef`",
                location=f"{cache_key}:spec.functionRef",
            )
        case registry.Resource() as watched_function:
            # Just needed to set `watched_function`
            pass

    function_under_test = get_resource_from_cache(
        resource_class=watched_function.resource_type, cache_key=watched_function.name
    )
    if not function_under_test:
        function_under_test = DepSkip(
            message=(
                f"{watched_function.resource_type.__name__} ({watched_function.name}) "
                f"not found or not ready, FunctionTest '{cache_key}' must wait."
            ),
            location=location,
        )

    initial_resource = spec.get("currentResource")

    cel_env = celpy.Environment(annotations=koreo_function_annotations)

    test_cases_spec = spec.get("testCases")
    test_cases = _prepare_test_cases(
        cel_env=cel_env,
        spec=test_cases_spec,
        resource_function=watched_function.resource_type == ResourceFunction,
    )

    if is_error(test_cases):
        return PermFail(
            message=test_cases.message,
            location=f"FunctionTest:{cache_key}:{test_cases.location}",
        )

    raw_inputs = spec.get("inputs", {})
    if not raw_inputs:
        inputs = None
    else:
        inputs = celpy.json_to_cel(raw_inputs)
        if not isinstance(inputs, celtypes.MapType):
            return PermFail(
                message=f"FunctionTest '{cache_key}' `spec.inputs` ('{inputs}') must be an object.",
                location=location,
            )

    watched_resources: list[registry.Resource] = [watched_function]

    resource_template_resources = _check_for_resource_template_ref(
        function_under_test, inputs, test_cases
    )
    if resource_template_resources:
        watched_resources.extend(resource_template_resources)

    return (
        structure.FunctionTest(
            function_under_test=function_under_test,
            inputs=inputs,
            test_cases=test_cases,
            initial_resource=initial_resource,
        ),
        watched_resources,
    )


def _prepare_test_cases(
    cel_env: celpy.Environment, spec: list[dict] | None, resource_function: bool
) -> Sequence[structure.TestCase] | PermFail:
    test_cases = []

    if not spec:
        return test_cases

    for idx, test_spec in enumerate(spec):
        test_case = _prepare_test_case(
            cel_env, spec=test_spec, idx=idx, resource_function=resource_function
        )
        if is_error(test_case):
            return test_case

        test_cases.append(test_case)

    return test_cases


def _prepare_test_case(
    cel_env: celpy.Environment, spec: dict, idx: int, resource_function: bool
) -> structure.TestCase | PermFail:
    variant = spec.get("variant", False)

    label = spec.get("label")
    if not label:
        label = f"Test Case {idx + 1}"

    if variant and "variant" not in label.lower():
        label = f"[variant] {label}"

    location = f'spec.testCases[{idx}: "{label}"]'

    # This screwy structure is just due to Python's crappy type narrowing.
    raw_overrides = spec.get("inputOverrides", {})
    if not raw_overrides:
        overrides = None
    else:
        overrides = celpy.json_to_cel(raw_overrides)
        if not isinstance(overrides, celtypes.MapType):
            return PermFail(
                message=f"`{location}.inputOverrides` must be an object.",
                location=f"{location}.inputOverrides",
            )

    current_resource = spec.get("currentResource")
    if current_resource is not None and not isinstance(current_resource, dict):
        return PermFail(
            message=f"`{location}.currentResource` must be an object.",
            location=f"{location}.currentResource",
        )

    overlay_resource = spec.get("overlayResource")
    if overlay_resource is not None and not isinstance(overlay_resource, dict):
        return PermFail(
            message=f"`{location}.overlayResource` must be an object.",
            location=f"{location}.overlayResource",
        )

    match prepare_overlay_expression(
        cel_env=cel_env,
        spec=overlay_resource,
        location=f"{location}.overlayResource",
    ):
        case PermFail() as err:
            return err
        case None:
            overlay = None
        case _ as overlay:
            pass

    # Just used so that there is only one.
    bad_assertions_failure = PermFail(
        message=(
            f"`{location}` must contain exactly one assertion (`expectResource`, `expectReturn`, `expectOutcome`, or `expectDelete`)."
        ),
        location=f"{location}",
    )

    assertion = None
    expected_resource = spec.get("expectResource")
    if expected_resource is not None:
        if not resource_function:
            return PermFail(
                message=f"`{location}.expectResource` only valid for `ResourceFunction` tests.",
                location=f"{location}.expectResource",
            )

        if not isinstance(expected_resource, dict):
            return PermFail(
                message=f"`{location}.expectResource` must be an object.",
                location=f"{location}.expectResource",
            )

        if not expected_resource:
            return PermFail(
                message=f"`{location}.expectResource` may not be an empty object.",
                location=f"{location}.expectResource",
            )

        if assertion:
            return bad_assertions_failure

        assertion = structure.ExpectResource(expected_resource)

    expected_outcome_spec = spec.get("expectOutcome")
    if expected_outcome_spec is not None:
        if not isinstance(expected_outcome_spec, dict):
            return PermFail(
                message=(f"`{location}.expectOutcome` must be an object."),
                location=f"{location}.expectOutcome",
            )

        if not expected_outcome_spec:
            return PermFail(
                message=(f"`{location}.expectOutcome` must be a valid outcome."),
                location=f"{location}.expectOutcome",
            )

        if assertion:
            return bad_assertions_failure

        expected_outcome_spec["assert"] = True
        assertion = structure.ExpectOutcome(
            predicate_to_koreo_result(
                celpy.json_to_cel([expected_outcome_spec]),
                location=f"{location}.expectOutcome",
            )
        )

    expected_return = spec.get("expectReturn")
    if expected_return is not None:
        if not isinstance(expected_return, dict):
            return PermFail(
                message=(f"`{location}.expectReturn` must be an object."),
                location=f"{location}.expectReturn",
            )

        if not expected_return:
            return PermFail(
                message=(f"`{location}.expectReturn` may not be an empty object."),
                location=f"{location}.expectReturn",
            )

        if assertion:
            return bad_assertions_failure

        assertion = structure.ExpectReturn(expected_return)

    expected_delete = spec.get("expectDelete")
    if expected_delete is not None:
        if not resource_function:
            return PermFail(
                message=f"`{location}.expectDelete` only valid for `ResourceFunction` tests.",
                location=f"{location}.expectDelete",
            )

        if assertion:
            return bad_assertions_failure

        assertion = structure.ExpectDelete(expected_delete)

    if not assertion:
        return bad_assertions_failure

    return structure.TestCase(
        variant=variant,
        input_overrides=overrides,
        current_resource=current_resource,
        resource_overlay=overlay,
        assertion=assertion,
        label=label,
        skip=spec.get("skip", False),
    )


def _check_for_resource_template_ref(
    function_under_test, base_inputs, test_cases: Sequence[structure.TestCase]
) -> None | Sequence[registry.Resource[ResourceTemplate]]:
    if not is_unwrapped_ok(function_under_test):
        return None

    if not isinstance(function_under_test, ResourceFunction):
        return None

    if not isinstance(
        function_under_test.crud_config.resource_template, ResourceTemplateRef
    ):
        return None

    if not function_under_test.crud_config.resource_template.name:
        return None

    template_name = function_under_test.crud_config.resource_template.name
    name_args = extract_argument_structure(template_name.ast)

    # There are no inputs used, just compute the value.
    if not name_args:
        match evaluate(expression=template_name, inputs={}, location="templateName"):
            case celtypes.StringType() as name:
                return (registry.Resource(resource_type=ResourceTemplate, name=name),)
            case _:
                return None

    templates_used = set()
    for test_case in test_cases:
        if base_inputs and test_case.input_overrides:
            inputs = _overlay(base_inputs, test_case.input_overrides)
            if isinstance(inputs, celpy.CELEvalError):
                continue
        elif base_inputs:
            inputs = base_inputs
        elif test_case.input_overrides:
            inputs = test_case.input_overrides
        else:
            inputs = celtypes.MapType()

        if function_under_test.local_values:
            locals = function_under_test.local_values.evaluate(
                {
                    "inputs": inputs,
                }
            )
        else:
            locals = celtypes.MapType()

        match evaluate(
            expression=template_name,
            inputs={
                "locals": locals,
                "inputs": inputs,
            },
            location="templateName",
        ):
            case celtypes.StringType() as name:
                templates_used.add(name)
            case _:
                continue

    return [
        registry.Resource(resource_type=ResourceTemplate, name=name)
        for name in templates_used
    ]



import unittest

import celpy
from celpy import celtypes

from koreo import result
from koreo.cel.evaluation import evaluate_overlay
from koreo.value_function import prepare
from koreo.value_function.structure import ValueFunction


class TestValueFunctionPrepare(unittest.IsolatedAsyncioTestCase):
    async def test_full_spec(self):
        prepared = await prepare.prepare_value_function(
            cache_key="test",
            spec={
                "preconditions": [
                    {"assert": "=!inputs.skip", "skip": {"message": "Skip"}},
                    {
                        "assert": "=!inputs.permFail",
                        "permFail": {"message": "Perm Fail"},
                    },
                    {
                        "assert": "=!inputs.depSkip",
                        "depSkip": {"message": "Dep Skip"},
                    },
                ],
                "locals": {
                    "some_list": [1, 2, 3, 4],
                    "a_value_map": {"a": "b"},
                    "integer": 7,
                    "a_none": None,
                    "cel_expr": "=1 + 17 / 2",
                },
                "return": {
                    "string": "1 + 1",
                    "simple_cel": "=1 + 1",
                    "nested": {
                        "string": "this is a test",
                        "simple_cel": "='a fun' + ' test'",
                    },
                    "list": [
                        1,
                        2,
                        3,
                        "hopefully",
                        "it",
                        "works",
                        "=1 - 2",
                        "='a string' + ' concat'",
                    ],
                },
            },
        )

        function, _ = prepared
        self.assertIsInstance(
            function,
            ValueFunction,
        )

    async def test_missing_and_bad_spec(self):
        bad_specs = [None, {}, "asda", [], 2, True]
        for bad_spec in bad_specs:
            outcome = await prepare.prepare_value_function(
                cache_key="test", spec=bad_spec
            )
            self.assertIsInstance(
                outcome,
                result.PermFail,
                msg=f'Expected `PermFail` for malformed `spec` "{bad_spec}"',
            )

    async def test_malformed_locals(self):
        bad_locals = ["asda", [], 2, True]
        for bad_locals in bad_locals:
            outcome = await prepare.prepare_value_function(
                cache_key="test", spec={"locals": bad_locals}
            )
            self.assertIsInstance(
                outcome,
                result.PermFail,
                msg=f'Expected `PermFail` for malformed `spec.locals` "{bad_locals}"',
            )

    async def test_preconditions_none_and_empty_list(self):
        outcome = await prepare.prepare_value_function("test", {"preconditions": None})

        self.assertIsInstance(
            outcome,
            result.PermFail,
        )

        prepared = await prepare.prepare_value_function("test", {"preconditions": []})
        function, _ = prepared
        self.assertIsInstance(
            function,
            ValueFunction,
            msg="Unexpected error with empty list `spec.preconditions`",
        )

    async def test_bad_precondition_input_type(self):
        bad_values = [1, "abc", {"value": "one"}, True]
        for value in bad_values:
            self.assertIsInstance(
                await prepare.prepare_value_function("test", {"preconditions": value}),
                result.PermFail,
                msg=f"Expected PermFail for bad `predicate_spec` '{value}' (type: {type(value)})",
            )

    async def test_malformed_precondition_input(self):
        bad_values = [
            {"skip": {"message": "=1 + missing"}},
            {"assert": "=1 / 0 '", "permFail": {"message": "Bogus assert"}},
        ]
        self.assertIsInstance(
            await prepare.prepare_value_function("test", {"preconditions": bad_values}),
            result.PermFail,
        )

    async def test_none_and_empty_list_return(self):
        outcome = await prepare.prepare_value_function("test", {"return": None})
        self.assertIsInstance(outcome, result.PermFail)

        function, _ = await prepare.prepare_value_function("test", {"return": {}})
        self.assertIsInstance(
            function,
            ValueFunction,
            msg="Unexpected error with empty map `return`",
        )

    async def test_bad_return_input_type(self):
        bad_values = [1, "abc", ["value", "one"], True]
        for value in bad_values:
            self.assertIsInstance(
                await prepare.prepare_value_function("test", {"return": value}),
                result.PermFail,
                msg=f"Expected PermFail for bad `return` '{value}' (type: {type(value)})",
            )

    async def test_malformed_return_input(self):
        bad_values = {
            "skip": {"message": "=1 + missing"},
            "assert": "=1 / 0 '",
            "permFail": {"message": "Bogus assert"},
        }
        self.assertIsInstance(
            await prepare.prepare_value_function("test", {"return": bad_values}),
            result.PermFail,
        )

    async def test_ok_return_input(self):

        return_value_cel = {
            "string": "1 + 1",
            "simple_cel": "=1 + 1",
            "nested": {
                "string": "this is a test",
                "simple_cel": "='a fun' + ' test'",
            },
            "list": [
                1,
                2,
                3,
                "hopefully",
                "it",
                "works",
                "=1 - 2",
                "='a string' + ' concat'",
            ],
        }

        inputs = {
            "skip": celtypes.BoolType(False),
            "permFail": celtypes.BoolType(False),
            "depSkip": celtypes.BoolType(False),
            "retry": celtypes.BoolType(False),
            "ok": celtypes.BoolType(False),
        }

        expected_return = {
            "string": "1 + 1",
            "simple_cel": 2,
            "nested": {
                "string": "this is a test",
                "simple_cel": "a fun test",
            },
            "list": [
                1,
                2,
                3,
                "hopefully",
                "it",
                "works",
                -1,
                "a string concat",
            ],
        }

        prepared = await prepare.prepare_value_function(
            "test", {"return": return_value_cel}
        )

        assert result.is_unwrapped_ok(prepared)

        function, _ = prepared

        assert function.return_value is not None

        return_value = evaluate_overlay(
            overlay=function.return_value,
            inputs=inputs,
            base=celtypes.MapType({}),
            location="unittest",
        )
        self.assertDictEqual(
            return_value,
            expected_return,
        )



import copy
import json
import base64

from celpy import celtypes
import celpy


# NOTE: Annotations and Function List are at the bottom of the file. Sorry.


def _self_ref(
    resource: celtypes.MapType,
) -> celtypes.MapType | celpy.CELEvalError:

    api_version_key = celtypes.StringType("apiVersion")
    api_version = resource.get(api_version_key)
    if not api_version:
        return celpy.CELEvalError(f"Missing `apiVersion`.")

    kind_key = celtypes.StringType("kind")
    kind = resource.get(kind_key)
    if not kind:
        return celpy.CELEvalError(f"Missing `kind`.")

    metadata_key = celtypes.StringType("metadata")
    metadata = resource.get(metadata_key)
    if not metadata:
        return celpy.CELEvalError(f"Missing `metadata`.")

    name_key = celtypes.StringType("name")
    name = metadata.get(name_key)
    if not name:
        return celpy.CELEvalError(f"Missing `metadata.name`.")

    namespace_key = celtypes.StringType("namespace")
    namespace = metadata.get(namespace_key)
    if not namespace:
        return celpy.CELEvalError(f"Missing `metadata.namespace`.")

    reference = celtypes.MapType(
        {
            api_version_key: api_version,
            kind_key: kind,
            name_key: name,
            namespace_key: namespace,
        }
    )

    return reference


def _to_ref(
    source: celtypes.MapType,
) -> celtypes.MapType | celpy.CELEvalError:
    reference = celtypes.MapType()

    api_version_key = celtypes.StringType("apiVersion")
    api_version = source.get(api_version_key)
    if api_version:
        reference[api_version_key] = api_version

    kind_key = celtypes.StringType("kind")
    kind = source.get(kind_key)
    if kind:
        reference[kind_key] = kind

    if "external" in source:
        external_key = celtypes.StringType("external")
        external = source.get(external_key)
        if not external:
            return celpy.CELEvalError(f"`external` must contain a value.")

        reference[external_key] = external
        return reference

    if "name" not in source:
        return celpy.CELEvalError(
            f"`external` or `name` are required to build a reference."
        )

    name_key = celtypes.StringType("name")
    name = source.get(name_key)

    if not name:
        return celpy.CELEvalError(f"`name` must contain a value.")

    reference[name_key] = name

    namespace_key = celtypes.StringType("namespace")
    namespace = source.get(namespace_key)
    if namespace:
        reference[namespace_key] = namespace

    return reference


def _group_ref(
    source: celtypes.MapType,
) -> celtypes.MapType | celpy.CELEvalError:
    reference = celtypes.MapType()

    api_group_key = celtypes.StringType("apiGroup")
    api_group = source.get(api_group_key)
    if api_group:
        reference[api_group_key] = api_group
    else:
        api_version_key = celtypes.StringType("apiVersion")
        api_version = source.get(api_version_key)
        if api_version and isinstance(api_version, celtypes.StringType):
            parts = api_version.split("/", 1)
            api_group = parts[0]
            reference[api_group_key] = celtypes.StringType(api_group)

    kind_key = celtypes.StringType("kind")
    kind = source.get(kind_key)
    if kind:
        reference[kind_key] = kind

    if "external" in source:
        external_key = celtypes.StringType("external")
        external = source.get(external_key)
        if not external:
            return celpy.CELEvalError(f"`external` must contain a value.")

        reference[external_key] = external
        return reference

    if "name" not in source:
        return celpy.CELEvalError(
            f"`external` or `name` are required to build a reference."
        )

    name_key = celtypes.StringType("name")
    name = source.get(name_key)

    if not name:
        return celpy.CELEvalError(f"`name` must contain a value.")

    reference[name_key] = name

    namespace_key = celtypes.StringType("namespace")
    namespace = source.get(namespace_key)
    if namespace:
        reference[namespace_key] = namespace

    return reference


def _kindless_ref(
    source: celtypes.MapType,
) -> celtypes.MapType | celpy.CELEvalError:
    reference = celtypes.MapType()

    if "external" in source:
        external_key = celtypes.StringType("external")
        external = source.get(external_key)
        if not external:
            return celpy.CELEvalError(f"`external` must contain a value.")

        reference[external_key] = external
        return reference

    if "name" not in source:
        return celpy.CELEvalError(
            f"`external` or `name` are required to build a reference."
        )

    name_key = celtypes.StringType("name")
    name = source.get(name_key)

    if not name:
        return celpy.CELEvalError(f"`name` must contain a value.")

    reference[name_key] = name

    namespace_key = celtypes.StringType("namespace")
    namespace = source.get(namespace_key)
    if namespace:
        reference[namespace_key] = namespace

    return reference


def _config_connect_ready(
    resource: celtypes.MapType,
) -> celtypes.BoolType | celpy.CELEvalError:
    if "status" not in resource:
        return celtypes.BoolType(False)

    status: celtypes.MapType = resource.get("status")

    if "conditions" not in status:
        return celtypes.BoolType(False)

    conditions: celtypes.ListType = status.get("conditions")

    ready_condition = None

    for condition in conditions:
        condition_type = condition.get("type")
        if condition_type != "Ready":
            continue

        # This is done to ensure there is only one 'type=Ready' I am unsure if
        # there is a valid case for multiple 'type=Ready' conditions existing.
        # If there is, we'll need to inspect it and decide how to handle it.
        if ready_condition is not None:
            return celtypes.BoolType(False)

        ready_condition = condition

    if not ready_condition:
        return celtypes.BoolType(False)

    reason = ready_condition.get("reason")

    if reason != "UpToDate":
        return celtypes.BoolType(False)

    condition_status = ready_condition.get("status")

    if condition_status != "True":
        return celtypes.BoolType(False)

    return celtypes.BoolType(True)


def _overlay(
    resource: celtypes.MapType,
    overlay: celtypes.MapType,
) -> celtypes.MapType | celpy.CELEvalError:
    return _deep_overlay(resource, overlay)


def _deep_overlay(
    resource: celtypes.MapType,
    overlay: celtypes.MapType,
) -> celtypes.MapType | celpy.CELEvalError:
    resource = copy.deepcopy(resource)

    for field, overlay_value in overlay.items():
        if field in resource:
            resource_value = resource[field]
            if isinstance(overlay_value, celtypes.MapType) and isinstance(
                resource_value, celtypes.MapType
            ):
                updated_value = _deep_overlay(resource_value, overlay_value)

                if isinstance(updated_value, celpy.CELEvalError):
                    return updated_value

                resource[field] = updated_value

                continue

        resource[field] = overlay_value

    return resource


def _flatten(resource: celtypes.ListType) -> celtypes.ListType | celpy.CELEvalError:
    if not resource:
        return celtypes.ListType()

    merged = celtypes.ListType()

    for nested in resource:
        merged.extend(nested)

    return merged


def _lower(string: celtypes.StringType) -> celtypes.StringType | celpy.CELEvalError:
    return celtypes.StringType(string.lower())


def _strip(string: celtypes.StringType, on: celtypes.StringType) -> celtypes.StringType:
    return celtypes.StringType(string.strip(on))


def _rstrip(
    string: celtypes.StringType, on: celtypes.StringType
) -> celtypes.StringType:
    return celtypes.StringType(string.rstrip(on))


def _split(
    string: celtypes.StringType, on: celtypes.StringType
) -> celtypes.ListType | celpy.CELEvalError:
    if not on:
        return celpy.CELEvalError(f"split separator may not be empty")

    return celtypes.ListType(celtypes.StringType(part) for part in string.split(on))


def _split_first(
    string: celtypes.StringType, on: celtypes.StringType
) -> celtypes.StringType | celpy.CELEvalError:
    if not on:
        return celpy.CELEvalError(f"split separator may not be empty")

    if not string:
        return celtypes.StringType("")

    return celtypes.StringType(string.split(on, 1)[0])


def _split_last(
    string: celtypes.StringType, on: celtypes.StringType
) -> celtypes.StringType | celpy.CELEvalError:
    if not on:
        return celpy.CELEvalError(f"split separator may not be empty")

    if not string:
        return celtypes.StringType("")

    return celtypes.StringType(string.rsplit(on, 1)[-1])


def _split_index(
    string: celtypes.StringType, on: celtypes.StringType, index: celtypes.IntType
) -> celtypes.StringType | celpy.CELEvalError:
    if not on:
        return celpy.CELEvalError(f"split separator may not be empty")

    if not string:
        return celtypes.StringType("")

    try:
        split = string.split(on)[index]
    except IndexError:
        return celpy.CELEvalError(f"index out of bounds on split")

    return celtypes.StringType(split)


def _to_json(value: celtypes.Value) -> celtypes.StringType | celpy.CELEvalError:
    try:
        return celtypes.StringType(json.dumps(value))
    except Exception as err:
        return celpy.CELEvalError(f"JSON Encoding error {err}")


def _from_json(value: celtypes.StringType) -> celtypes.Value | celpy.CELEvalError:
    try:
        return celpy.json_to_cel(json.loads(value))
    except Exception as err:
        return celpy.CELEvalError(f"JSON Decoding error {err}")


def _b64encode(value: celtypes.Value) -> celtypes.StringType | celpy.CELEvalError:
    try:
        return celtypes.StringType(base64.b64encode(value.encode()).decode())
    except Exception as err:
        return celpy.CELEvalError(f"Base64 Encoding error {err}")


def _b64decode(value: celtypes.Value) -> celtypes.StringType | celpy.CELEvalError:
    try:
        return celtypes.StringType(base64.b64decode(value))
    except Exception as err:
        return celpy.CELEvalError(f"Base64 Encoding error {err}")


koreo_function_annotations: dict[str, celpy.Annotation] = {
    "to_ref": celtypes.FunctionType,
    "self_ref": celtypes.FunctionType,
    "group_ref": celtypes.FunctionType,
    "kindless_ref": celtypes.FunctionType,
    "config_connect_ready": celtypes.FunctionType,
    "overlay": celtypes.FunctionType,
    "lower": celtypes.FunctionType,
    "flatten": celtypes.FunctionType,
    "split": celtypes.FunctionType,
    "split_first": celtypes.FunctionType,
    "split_last": celtypes.FunctionType,
    "split_index": celtypes.FunctionType,
    "strip": celtypes.FunctionType,
    "rstrip": celtypes.FunctionType,
    "to_json": celtypes.FunctionType,
    "from_json": celtypes.FunctionType,
    "b64encode": celtypes.FunctionType,
    "b64decode": celtypes.FunctionType,
}

koreo_cel_functions: dict[str, celpy.CELFunction] = {
    "to_ref": _to_ref,
    "self_ref": _self_ref,
    "group_ref": _group_ref,
    "kindless_ref": _kindless_ref,
    "config_connect_ready": _config_connect_ready,
    "overlay": _overlay,
    "lower": _lower,
    "flatten": _flatten,
    "split": _split,
    "split_first": _split_first,
    "split_last": _split_last,
    "split_index": _split_index,
    "strip": _strip,
    "rstrip": _rstrip,
    "to_json": _to_json,
    "from_json": _from_json,
    "b64encode": _b64encode,
    "b64decode": _b64decode,
}



from typing import Awaitable, Callable, NamedTuple, Sequence, TypeVar
import asyncio
import copy
import logging
import time

from koreo.constants import ACTIVE_LABEL
from koreo.result import UnwrappedOutcome, is_unwrapped_ok
from koreo import registry


logger = logging.getLogger(name="koreo.cache")

T = TypeVar("T")
type PreparerFn[T] = Callable[
    [str, dict],
    Awaitable[UnwrappedOutcome[tuple[T, Sequence[registry.Resource] | None]]],
]


def get_resource_from_cache(
    resource_class: type[T], cache_key: str
) -> UnwrappedOutcome[T] | None:
    resource_class_name = resource_class.__qualname__
    if not __CACHE.get(resource_class_name):
        __CACHE[resource_class_name] = {}

    cached = __CACHE[resource_class_name].get(cache_key)

    if cached:
        return cached.resource

    return None


class __CachedResource[T](NamedTuple):
    spec: dict
    resource: UnwrappedOutcome[T]
    resource_version: str
    prepared_at: float | None = None
    system_data: dict | None = None


def get_resource_system_data_from_cache(
    resource_class: type[T], cache_key: str
) -> __CachedResource[T] | None:
    resource_class_name = resource_class.__qualname__
    if not __CACHE.get(resource_class_name):
        __CACHE[resource_class_name] = {}

    return __CACHE[resource_class_name].get(cache_key)


async def prepare_and_cache(
    resource_class: type[T],
    preparer: PreparerFn[T],
    metadata: dict,
    spec: dict,
    _system_data: dict | None = None,
) -> UnwrappedOutcome[T]:
    resource_metadata = _extract_meta(metadata=metadata)

    cache_key = resource_metadata.resource_name

    resource_class_name = resource_class.__qualname__

    if not __CACHE.get(resource_class_name):
        __CACHE[resource_class_name] = {}

    cached = __CACHE[resource_class_name].get(cache_key)

    if cached and cached.resource_version == resource_metadata.resource_version:
        return cached.resource

    prepare_started_at = time.monotonic()
    resource = registry.Resource(resource_type=resource_class, name=cache_key)
    registry.register(registerer=resource)
    preparer_outcome = await preparer(cache_key, copy.deepcopy(spec))
    prepare_finished_at = time.monotonic()

    if is_unwrapped_ok(preparer_outcome):
        prepared_resource, subscriptions = preparer_outcome
    else:
        prepared_resource = preparer_outcome
        subscriptions = None

    __CACHE[resource_class_name][cache_key] = __CachedResource[T](
        spec=spec,
        resource=prepared_resource,
        resource_version=resource_metadata.resource_version,
        prepared_at=prepare_started_at,
        system_data=_system_data,
    )
    logger.debug(
        f"Updating {resource_class_name} cache for {cache_key} ({resource_metadata.resource_version})."
    )

    await _handle_notifications(
        resource=resource,
        preparer=preparer,
        subscriptions=subscriptions,
        prepare_started_at=prepare_started_at,
        prepare_finished_at=prepare_finished_at,
    )

    return prepared_resource


async def delete_resource_from_cache(
    resource_class: type[T],
    metadata: dict,
) -> None:
    resource_metadata = _extract_meta(metadata=metadata)
    cache_key = resource_metadata.resource_name

    await delete_from_cache(
        resource_class=resource_class,
        cache_key=cache_key,
    )


async def delete_from_cache(
    resource_class: type[T], cache_key: str, version: str | None = None
) -> None:
    resource_class_name = resource_class.__qualname__
    if not __CACHE.get(resource_class_name):
        return None

    cached = __CACHE[resource_class_name].get(cache_key)
    if not cached:
        return None

    if version and version != cached.resource_version:
        logger.debug(
            f"Skip deleting {cache_key} from {resource_class_name} cache due to version mismatch ({version} != {cached.resource_version})."
        )
        return None

    deleted_at = time.monotonic()
    del __CACHE[resource_class_name][cache_key]

    resource = registry.Resource(resource_type=resource_class, name=cache_key)
    queue = registry.kill_resource(resource=resource)
    registry.deregister(deregisterer=resource, deregistered_at=deleted_at)

    # This shouldn't happen, just an extra check
    if resource in _REPREPARE_TASKS:
        _REPREPARE_TASKS[resource].cancel()

    match queue:
        case None:
            pass
        case asyncio.Queue():
            async with asyncio.timeout(5):
                try:
                    await queue.join()
                except asyncio.TimeoutError:
                    pass

    logger.debug(f"Deleted {cache_key} from {resource_class_name} cache.")

    return None


_PREPARE_TIMES: dict[registry.Resource, float] = {}


async def _handle_notifications(
    resource: registry.Resource[T],
    subscriptions: Sequence[registry.Resource] | None,
    prepare_started_at: float,
    prepare_finished_at: float,
    preparer: PreparerFn[T] | None = None,
):
    # We need to track the earliest possible time of prepare.
    _PREPARE_TIMES[resource] = prepare_started_at

    if not subscriptions:
        subscriptions = []

    registry.subscribe_only_to(subscriber=resource, resources=subscriptions)

    # We need to send the latest possible time to subscribers.
    registry.notify_subscribers(notifier=resource, event_time=prepare_finished_at)

    # If this resource watches other resources, setup the monitor to reprepare
    if subscriptions and preparer and resource not in _REPREPARE_TASKS:
        task_name = f"{resource.resource_type.__qualname__}:{resource.name}"
        resource_task = asyncio.create_task(
            _monitor_and_reprepare(
                resource=resource,
                preparer=preparer,
            ),
            name=task_name,
        )
        _REPREPARE_TASKS[resource] = resource_task
        resource_task.add_done_callback(_deletor(resource))


async def _monitor_and_reprepare(
    resource: registry.Resource[T],
    preparer: PreparerFn[T],
):
    queue = registry.register(registerer=resource)
    caught_exception = None
    while True:
        try:
            event = await queue.get()
        except (asyncio.CancelledError, asyncio.QueueShutDown):
            break

        try:
            match event:
                case registry.Kill():
                    break
                case (_, event_time) if event_time <= _PREPARE_TIMES[resource]:
                    continue
                case (_, event_time):
                    try:
                        await _reprepare_and_update_cache(
                            resource_class=resource.resource_type,
                            cache_key=resource.name,
                            preparer=preparer,
                        )
                    except Exception as err:
                        logger.error(
                            f"Critical error preparing {resource.name} "
                            f"({resource.resource_type.__qualname__}) {err}"
                        )
                        caught_exception = err
                        break

        finally:
            queue.task_done()

    if caught_exception:
        raise caught_exception


async def _reprepare_and_update_cache(
    resource_class: type[T],
    preparer: PreparerFn[T],
    cache_key: str,
) -> None:
    resource_class_name = resource_class.__qualname__

    if not __CACHE.get(resource_class_name):
        return

    cached = __CACHE[resource_class_name].get(cache_key)
    if not cached:
        return

    prepare_started_at = time.monotonic()
    preparer_outcome = await preparer(cache_key, copy.deepcopy(cached.spec))
    prepare_finished_at = time.monotonic()

    if is_unwrapped_ok(preparer_outcome):
        prepared_resource, subscriptions = preparer_outcome
    else:
        prepared_resource = preparer_outcome
        subscriptions = None

    __CACHE[resource_class_name][cache_key] = cached._replace(
        resource=prepared_resource, prepared_at=prepare_started_at
    )

    logger.debug(
        f"Repreparing {cache_key} ({cached.resource_version}) in {resource_class_name} cache."
    )

    resource = registry.Resource(resource_type=resource_class, name=cache_key)
    await _handle_notifications(
        resource=resource,
        subscriptions=subscriptions,
        prepare_started_at=prepare_started_at,
        prepare_finished_at=prepare_finished_at,
    )


_REPREPARE_TASKS: dict[registry.Resource, asyncio.Task] = {}


def _deletor(resource):
    def do_delete(task: asyncio.Task):
        del _REPREPARE_TASKS[resource]
        registry.deregister(resource, time.monotonic())

        task_exception = task.exception()
        if task_exception:
            logger.error(f"Re-preparer for {resource} failed with {task_exception}")
            print(f"Re-preparer for {resource} failed with {task_exception}")

    return do_delete


__CACHE: dict[str, dict[str, __CachedResource]] = {}


class __ResourceMetadata(NamedTuple):
    resource_name: str
    resource_version: str

    active: bool


def _extract_meta(metadata: dict) -> __ResourceMetadata:
    resource_name = metadata.get("name")
    resource_version = metadata.get("resourceVersion")

    if not (resource_name and resource_version):
        raise TypeError("Bad Resource: resource name and version are required.")

    labels = metadata.get("labels", {})

    label_active = labels.get(ACTIVE_LABEL, "true").lower() in (
        "t",
        "true",
    )

    return __ResourceMetadata(
        resource_name=resource_name,
        resource_version=resource_version,
        active=label_active,
    )


def _reset_cache():
    """This is for unit testing."""
    __CACHE.clear()

    for task in _REPREPARE_TASKS.values():
        task.cancel()

    registry._reset_registries()
